//
//  PcmPlayer.m
//  TXCopyrightedMediaDemo
//
//  Created by chavezchen on 2022/6/15.
//

#import "PcmPlayer.h"

static int const kMinSizePerFrame = 20480; // 每帧最小数据长度

@interface PcmPlayer ()
{
    AudioStreamBasicDescription audioDescription;
    AudioQueueRef audioQueue;//音频播放队列
    AudioQueueBufferRef audioQueueBuffers[kBufferSize];//音频缓存
    BOOL audioQueueUsed[kBufferSize]; // 缓存buffer是否被使用
}
@property (nonatomic, strong) dispatch_queue_t playQueue;
@property (nonatomic, assign) BOOL isInited;
@property (nonatomic, assign) BOOL isPrepared;
@property (atomic, assign) BOOL isPlaying;
@property (nonatomic, strong) NSMutableArray *cacheDataArray;
@property (nonatomic, assign) NSInteger enQueueCount;
// pcm每秒数据量,用来计算时长
@property (nonatomic, assign) NSInteger bytesPerSecond;
//@property (nonatomic, assign) float playedTime; // 已经播放的时长
@end


@implementation PcmPlayer

- (instancetype)initWithSampleRate:(NSInteger)sampleRate channelsPerFrame:(int)channelsPerFrame bitsPerChannel:(int)bitsPerChannel
{
    if (self == [super init]) {
        self.playQueue = dispatch_queue_create("audio play queue", DISPATCH_QUEUE_SERIAL);
        audioDescription.mSampleRate = sampleRate;//采样率
        audioDescription.mFormatID = kAudioFormatLinearPCM;
        audioDescription.mFormatFlags = (kAudioFormatFlagIsSignedInteger | kAudioFormatFlagIsPacked);
        audioDescription.mChannelsPerFrame = channelsPerFrame; // 声道
        audioDescription.mFramesPerPacket = 1; // 每一个packet一侦数据
        audioDescription.mBitsPerChannel = bitsPerChannel;
        audioDescription.mBytesPerFrame = (audioDescription.mBitsPerChannel / 8) * audioDescription.mChannelsPerFrame;
        audioDescription.mBytesPerPacket = audioDescription.mBytesPerFrame * audioDescription.mFramesPerPacket;
        // 每秒数据量
        self.bytesPerSecond = audioDescription.mSampleRate * audioDescription.mBitsPerChannel / 8 * audioDescription.mChannelsPerFrame;

        [self setupAudioQueue];
    }
    return self;
}

- (void)updateTrack:(NSInteger)sampleRate channelCount:(int)channelCount
{
    [self pause];
    if (audioQueue != nil) {
        AudioQueueStop(audioQueue,true);
        AudioQueueReset(audioQueue);
    }
    [self.cacheDataArray removeAllObjects];
    audioDescription.mChannelsPerFrame = channelCount;
    audioDescription.mSampleRate = sampleRate;
    // 每秒数据量
    self.bytesPerSecond = audioDescription.mSampleRate * audioDescription.mBitsPerChannel / 8 * audioDescription.mChannelsPerFrame;
    
    [self setupAudioQueue];
}

- (void)setupAudioQueue
{
    OSStatus status = AudioQueueNewOutput(&audioDescription,AudioPlayerAQInputCallback, (__bridge void*)self,nil,nil,0, &audioQueue); //使用player的内部线程播放
    if (status != 0) {
        NSLog(@"AudioQueueNewOutput start failed!==%d",(int)status);
        return;
    }
    // 初始化音频缓冲区
    for (int i =0; i < kBufferSize; i++) {
        int result =AudioQueueAllocateBuffer(audioQueue,kMinSizePerFrame, &audioQueueBuffers[i]);
        NSLog(@"AudioQueueAllocateBuffer i = %d,result = %d", i, result);
    }
    self.isInited = YES;
}

static void AudioPlayerAQInputCallback(void* inUserData,AudioQueueRef outQ, AudioQueueBufferRef outQB) {
    PcmPlayer *player = (__bridge PcmPlayer*)inUserData;
    if (player == nil) {
        return;
    }
    [player playerCallback:outQB];
}

- (void)playerCallback:(AudioQueueBufferRef)outQB {
    dispatch_sync(self.playQueue, ^{
        @synchronized (self) {
            for (int i =0; i < kBufferSize; i++) {
                if (outQB == audioQueueBuffers[i]) {
                    audioQueueUsed[i] = NO;
                }
            }
        }
        
        [self enqueueBuffers];
        if ([self.delegate respondsToSelector:@selector(pcmPlayerPlayedFrameData)]) {
            [self.delegate pcmPlayerPlayedFrameData];
        }
    });
    
}

- (void)play
{
    if(self.isPlaying)
    {
        NSLog(@"pcm player: the player is playing");
        return;
    }
    
    if (!self.isInited) {
        NSLog(@"pcm player: start audio player failed, not init !");
    }
    OSStatus status = AudioQueueStart(audioQueue, NULL);
    if (status != 0) {
        NSLog(@"pcm player: start audio player failed, %d!",status);
        return;
    }
    self.isPlaying = YES;
    if (self.isPrepared) {
        return;
    }
    // 第一次播放需要初始化一下缓冲区
    [self prepare];
}

- (void)prepare
{
    if (self.cacheDataArray.count < kBufferSize) {
        NSLog(@"pcm player: please push enough data. call: [pcmplayer pushData:]");
        return;
    }
    dispatch_async(self.playQueue, ^{
        // 先enqueue kBufferSize个buffer 初始化一下缓冲区
        for (int i = 0; i < kBufferSize; i++) {
            [self enqueueBuffers];
        }
    });
    self.isPrepared = YES;
}

- (void)enqueueBuffers
{
    if (self.cacheDataArray.count == 0) {
        NSLog(@"pcm player: not enough data");
        return;
    }
    
    // 从缓冲区取未使用的buffer
    AudioQueueBufferRef audioQueueBuffer = NULL;
    audioQueueBuffer = [self getNotUsedBuffer];
    if (audioQueueBuffer == NULL) {
        NSLog(@"pcm player: not find free buffer！");
        return;
    }
    
    __block NSData *sourceData = nil;
    @synchronized (self) {
        sourceData = self.cacheDataArray.firstObject;
        if (sourceData) {
            [self.cacheDataArray removeObjectAtIndex:0];
        }
        NSLog(@"pcm player: cacheDataArray count : %zd",self.cacheDataArray.count);
    }
    
    audioQueueBuffer->mAudioDataByteSize = sourceData.length;

    Byte *audioData = (Byte *)audioQueueBuffer->mAudioData;
    void *data = (void *)[sourceData bytes];
    memcpy(audioData, data, sourceData.length);
    
    OSStatus status = AudioQueueEnqueueBuffer(audioQueue, audioQueueBuffer, 0, NULL);
    if (status != noErr) {
        NSLog(@"pcm player: EnqueueBufferFailed! status: %d",(int)status);
    } else {
        float time = (float)sourceData.length / self.bytesPerSecond;
        self.playedTime += time;
        [self callBackPlayTime];
        NSLog(@"play dataSize:%zd",sourceData.length);
    }
}

- (AudioQueueBufferRef)getNotUsedBuffer {
    @synchronized (self) {
        for (int i =0; i < kBufferSize; i++) {
            if (NO == audioQueueUsed[i]) {
                audioQueueUsed[i] = YES;
                return audioQueueBuffers[i];
            }
        }
    }
    return NULL;
}

- (void)pushData:(NSData *)data
{
    dispatch_async(self.playQueue, ^{
        @synchronized (self) {
            if (data) {
                [self.cacheDataArray addObject:data];
                
//                if (self.enQueueCount < kBufferSize && self.isPlaying) {
//                    [self enqueueBuffers];
//                    self.enQueueCount++;
//                }
            } else {
                NSLog(@"pcm player: pushed empty data!");
            }
            
        }
    });
}

- (void)pause
{
    if (!self.isPlaying) {
        return;
    }
    self.isPlaying = NO;
    AudioQueuePause(audioQueue);
}

- (void)stop
{
    if (audioQueue != nil) {
        AudioQueueStop(audioQueue,true);
        AudioQueueReset(audioQueue);
    }
    audioQueue = nil;
    self.isPlaying = NO;
    self.isPrepared = NO;
    self.playedTime = 0;
}

- (void)seek:(int)pts
{
    AudioQueuePause(audioQueue);
    self.playedTime = pts;
    self.isPlaying = NO;
    self.isPrepared = NO;
    @synchronized (self) {
        [self.cacheDataArray removeAllObjects];
    }
    if ([self.delegate respondsToSelector:@selector(pcmPlayerSeeked:)]) {
        [self.delegate pcmPlayerSeeked:pts];
    }
}

- (NSMutableArray *)cacheDataArray
{
    if (!_cacheDataArray) {
        _cacheDataArray = [NSMutableArray array];
    }
    return _cacheDataArray;
}

- (void)dealloc
{
    if (audioQueue) {
        AudioQueueStop(audioQueue, YES);
    }
    audioQueue = nil;
}

- (void)callBackPlayTime
{
    dispatch_async(dispatch_get_main_queue(), ^{
        if ([self.delegate respondsToSelector:@selector(pcmPlayerPlayedTime:)]) {
            [self.delegate pcmPlayerPlayedTime:self.playedTime];
        }
    });
}


@end
