//
//  TXCopyrightHTTPManager.m
//  TXCopyrightedMediaDemo
//
//  Created by Zhirui Ou on 2021/5/19.
//

#import "TXCopyrightHTTPManager.h"

typedef void (^TXCopyrightURLSessionTaskCompletionHandler)(NSURLResponse *response, id responseObject, NSError *error);

@interface TXCopyrightHTTPManagerTaskInfo : NSObject

@property (atomic, strong) NSURLSessionDataTask *task;
@property (atomic, strong) NSThread *thread;
@property (nonatomic, strong) NSMutableData *mutableData;
@property (atomic, strong) id<NSURLSessionDataDelegate> delegate;
@property (nonatomic, copy) TXCopyrightURLSessionTaskCompletionHandler completionHandle;

- (instancetype)initWithTask:(NSURLSessionDataTask *)task completionHandle:(TXCopyrightURLSessionTaskCompletionHandler)completionHandle;
- (void)performBlock:(dispatch_block_t)block;
- (void)invalidate;

@end

@implementation TXCopyrightHTTPManagerTaskInfo

- (instancetype)initWithTask:(NSURLSessionDataTask *)task completionHandle:(TXCopyrightURLSessionTaskCompletionHandler)completionHandle
{
    if (self = [super init]) {
        self.mutableData = [NSMutableData data];
        self.task = task;
        self.completionHandle = completionHandle;
    }
    return self;
}

- (void)performBlock:(dispatch_block_t)block
{
    [self performSelector:@selector(performBlockOnClientThread:) onThread:self.thread withObject:[block copy] waitUntilDone:NO modes:@[NSDefaultRunLoopMode]];
}

- (void)performBlockOnClientThread:(dispatch_block_t)block
{
    block();
}

- (void)invalidate
{
    self.delegate = nil;
    self.thread = nil;
    self.mutableData = [NSMutableData data];
}

@end


#pragma mark - TXCAdapterHTTPManager
@interface TXCopyrightHTTPManager ()<NSURLSessionDelegate>
@property (atomic, strong, readonly) NSMutableDictionary *taskInfoByTaskID;
@property (atomic, strong, readonly) NSOperationQueue *sessionDelegateQueue;
@property (atomic, copy, readonly) NSURLSessionConfiguration *configuration;
@property (atomic, strong, readonly) NSURLSession *session;
@end

@implementation TXCopyrightHTTPManager

+ (TXCopyrightHTTPManager *)manager
{
    static dispatch_once_t      onceToken;
    static TXCopyrightHTTPManager *manager;
    dispatch_once(&onceToken, ^{
        NSURLSessionConfiguration *configuration = [NSURLSessionConfiguration defaultSessionConfiguration];
        manager = [[TXCopyrightHTTPManager alloc] initWithConfiguration:configuration];
    });
    return manager;
}

- (instancetype)initWithConfiguration:(NSURLSessionConfiguration *)configuration
{
    if (self = [super init]) {
        if (configuration == nil) {
            configuration = [NSURLSessionConfiguration defaultSessionConfiguration];
        }
        self->_configuration = [configuration copy];
        self->_taskInfoByTaskID = [NSMutableDictionary dictionary];
        self->_sessionDelegateQueue = [[NSOperationQueue alloc] init];
        [self->_sessionDelegateQueue setMaxConcurrentOperationCount:1];
        [self->_sessionDelegateQueue setName:@"TXCAdapterHTTPManager"];
        // 设置session
        self->_session = [NSURLSession sessionWithConfiguration:configuration delegate:self delegateQueue:self->_sessionDelegateQueue];
    }
    return self;
}

- (NSURLSessionDataTask *)dataTaskWithRequest:(NSURLRequest *)request manager:(TXCopyrightHTTPManager *)manager success:(void (^)(NSURLSessionDataTask * _Nonnull, id _Nonnull))success failure:(void (^)(NSURLSessionDataTask * _Nonnull, NSError * _Nonnull))failure
{
    if (!request) {
        return nil;
    }
    NSURLSessionDataTask *task;
    TXCopyrightHTTPManagerTaskInfo *taskInfo;
    
    task = [self.session dataTaskWithRequest:request];
    taskInfo = [[TXCopyrightHTTPManagerTaskInfo alloc] initWithTask:task completionHandle:^(NSURLResponse *response, id responseObject, NSError *error) {
        dispatch_block_t callback = ^{
            if (error) {
                if (failure) {
                    failure(task, error);
                }
            } else {
//                id callBackObject = nil;
//                NSError *jsonError = nil;
//                callBackObject = [NSJSONSerialization JSONObjectWithData:responseObject options:NSJSONReadingMutableContainers error:&jsonError];
                if (success) {
                    success(task, responseObject);
                }
            }
        };
        if ([NSThread isMainThread]) {
            callback();
        } else {
            dispatch_async(dispatch_get_main_queue(), ^{
                callback();
            });
        }
    }];
    @synchronized (self) {
        self.taskInfoByTaskID[@(task.taskIdentifier)] = taskInfo;
    }
    return task;
}

- (TXCopyrightHTTPManagerTaskInfo *)taskInfoForTask:(NSURLSessionTask *)task
{
    
    TXCopyrightHTTPManagerTaskInfo *    result;
    @synchronized (self) {
        result = self.taskInfoByTaskID[@(task.taskIdentifier)];
    }
    return result;
}

- (NSURLSessionDataTask *)GET:(NSString *)URLString success:(void (^)(NSURLSessionDataTask * , id _Nonnull))success failure:(void (^)(NSURLSessionDataTask * _Nonnull, NSError * _Nonnull))failure
{
    if (!URLString || URLString.length <= 0) {
        return nil;
    }

    NSMutableURLRequest *mutableRequest = [NSMutableURLRequest requestWithURL:[NSURL URLWithString:URLString]];
    mutableRequest.HTTPMethod = @"GET";
    NSURLSessionDataTask *task = [self dataTaskWithRequest:mutableRequest manager:self success:success failure:failure];
    [task resume];
    
    return task;
}

+ (NSError *)errorWithErrorCode:(NSInteger)code description:(NSString *)description
{
    NSError *error = [NSError errorWithDomain:NSStringFromClass([self class])
                                         code:code
                                     userInfo:@{
                                         NSLocalizedDescriptionKey:description
                                     }
                      ];
    return error;
}

+ (NSString *)makeStringWithParamters:(NSDictionary*)parameters
{
    if (nil == parameters || [parameters count] == 0)
        return nil;
    
    NSMutableString* stringOfParamters = [[NSMutableString alloc] init];
    NSEnumerator *keyEnumerator = [parameters keyEnumerator];
    id key = nil;
    while ((key = [keyEnumerator nextObject])) {
        [stringOfParamters appendFormat:@"%@=%@&", key, [parameters valueForKey:key]];
    }
    // Delete last character of '&'
    NSRange lastCharRange = {[stringOfParamters length] - 1, 1};
    [stringOfParamters deleteCharactersInRange:lastCharRange];
    return stringOfParamters;
}

#pragma mark - <NSURLSessionDataDelegate>
// 将要重定向
- (void)URLSession:(NSURLSession *)session task:(NSURLSessionTask *)task willPerformHTTPRedirection:(NSHTTPURLResponse *)response newRequest:(NSURLRequest *)request completionHandler:(void (^)(NSURLRequest * _Nullable))completionHandler
{

    TXCopyrightHTTPManagerTaskInfo *taskInfo;
    taskInfo = [self taskInfoForTask:task];
    if ([taskInfo.delegate respondsToSelector:@selector(URLSession:task:willPerformHTTPRedirection:newRequest:completionHandler:)]) {
        [taskInfo performBlock:^{
            [taskInfo.delegate URLSession:session task:task willPerformHTTPRedirection:response newRequest:request completionHandler:completionHandler];
        }];
    } else {
        completionHandler(request);
    }
}
// 身份请求验证
- (void)URLSession:(NSURLSession *)session task:(NSURLSessionTask *)task didReceiveChallenge:(NSURLAuthenticationChallenge *)challenge completionHandler:(void (^)(NSURLSessionAuthChallengeDisposition disposition, NSURLCredential *credential))completionHandler
{
    TXCopyrightHTTPManagerTaskInfo *taskInfo;
    taskInfo = [self taskInfoForTask:task];
    if ([taskInfo.delegate respondsToSelector:@selector(URLSession:task:didReceiveChallenge:completionHandler:)]) {
        [taskInfo performBlock:^{
            [taskInfo.delegate URLSession:session task:task didReceiveChallenge:challenge completionHandler:completionHandler];
        }];
    } else {
        completionHandler(NSURLSessionAuthChallengePerformDefaultHandling, nil);
    }
}

- (void)URLSession:(NSURLSession *)session task:(NSURLSessionTask *)task needNewBodyStream:(void (^)(NSInputStream *bodyStream))completionHandler
{
    TXCopyrightHTTPManagerTaskInfo *    taskInfo;
    taskInfo = [self taskInfoForTask:task];
    if ([taskInfo.delegate respondsToSelector:@selector(URLSession:task:needNewBodyStream:)]) {
        [taskInfo performBlock:^{
            [taskInfo.delegate URLSession:session task:task needNewBodyStream:completionHandler];
        }];
    } else {
        completionHandler(nil);
    }
}

- (void)URLSession:(NSURLSession *)session task:(NSURLSessionTask *)task didSendBodyData:(int64_t)bytesSent totalBytesSent:(int64_t)totalBytesSent totalBytesExpectedToSend:(int64_t)totalBytesExpectedToSend
{
    TXCopyrightHTTPManagerTaskInfo *    taskInfo;

    taskInfo = [self taskInfoForTask:task];
    if ([taskInfo.delegate respondsToSelector:@selector(URLSession:task:didSendBodyData:totalBytesSent:totalBytesExpectedToSend:)]) {
        [taskInfo performBlock:^{
            [taskInfo.delegate URLSession:session task:task didSendBodyData:bytesSent totalBytesSent:totalBytesSent totalBytesExpectedToSend:totalBytesExpectedToSend];
        }];
    }
}

// 收到响应头
- (void)URLSession:(NSURLSession *)session dataTask:(NSURLSessionDataTask *)dataTask didReceiveResponse:(NSURLResponse *)response completionHandler:(void (^)(NSURLSessionResponseDisposition disposition))completionHandler
{
    TXCopyrightHTTPManagerTaskInfo *    taskInfo;
    
    taskInfo = [self taskInfoForTask:dataTask];
    if ([taskInfo.delegate respondsToSelector:@selector(URLSession:dataTask:didReceiveResponse:completionHandler:)]) {
        [taskInfo performBlock:^{
            [taskInfo.delegate URLSession:session dataTask:dataTask didReceiveResponse:response completionHandler:completionHandler];
        }];
    } else {
        completionHandler(NSURLSessionResponseAllow);
    }
}

// 接收数据
- (void)URLSession:(NSURLSession *)session dataTask:(NSURLSessionDataTask *)dataTask didReceiveData:(NSData *)data
{
    TXCopyrightHTTPManagerTaskInfo *    taskInfo;
    
    taskInfo = [self taskInfoForTask:dataTask];
    [taskInfo.mutableData appendData:data];
    if ([taskInfo.delegate respondsToSelector:@selector(URLSession:dataTask:didReceiveData:)]) {
        [taskInfo performBlock:^{
            [taskInfo.delegate URLSession:session dataTask:dataTask didReceiveData:data];
        }];
    }
}
// 请求完成，error有值代表失败
- (void)URLSession:(NSURLSession *)session task:(NSURLSessionTask *)task didCompleteWithError:(NSError *)error
{
    TXCopyrightHTTPManagerTaskInfo *    taskInfo;
    
    taskInfo = [self taskInfoForTask:task];
    if (error) {
        if (taskInfo.completionHandle) {
            taskInfo.completionHandle(nil, nil, error);
        }
    } else {
        if (taskInfo.completionHandle) {
            NSData *data = taskInfo.mutableData.copy;
            taskInfo.completionHandle(task.response, data, nil);
        }
    }
    
    @synchronized (self) {
        [self.taskInfoByTaskID removeObjectForKey:@(taskInfo.task.taskIdentifier)];
    }
    
    if ([taskInfo.delegate respondsToSelector:@selector(URLSession:task:didCompleteWithError:)]) {
        [taskInfo performBlock:^{
            [taskInfo.delegate URLSession:session task:task didCompleteWithError:error];
            [taskInfo invalidate];
        }];
    } else {
        [taskInfo invalidate];
    }
}
// 成为下载任务
- (void)URLSession:(NSURLSession *)session dataTask:(NSURLSessionDataTask *)dataTask didBecomeDownloadTask:(NSURLSessionDownloadTask *)downloadTask
{
    TXCopyrightHTTPManagerTaskInfo *    taskInfo;

    taskInfo = [self taskInfoForTask:dataTask];
    if ([taskInfo.delegate respondsToSelector:@selector(URLSession:dataTask:didBecomeDownloadTask:)]) {
        [taskInfo performBlock:^{
            [taskInfo.delegate URLSession:session dataTask:dataTask didBecomeDownloadTask:downloadTask];
        }];
    }
}

- (void)URLSession:(NSURLSession *)session dataTask:(NSURLSessionDataTask *)dataTask willCacheResponse:(NSCachedURLResponse *)proposedResponse completionHandler:(void (^)(NSCachedURLResponse *cachedResponse))completionHandler
{
    TXCopyrightHTTPManagerTaskInfo *    taskInfo;

    taskInfo = [self taskInfoForTask:dataTask];
    if ([taskInfo.delegate respondsToSelector:@selector(URLSession:dataTask:willCacheResponse:completionHandler:)]) {
        [taskInfo performBlock:^{
            [taskInfo.delegate URLSession:session dataTask:dataTask willCacheResponse:proposedResponse completionHandler:completionHandler];
        }];
    } else {
        completionHandler(proposedResponse);
    }
}

@end
